"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var clone_1 = tslib_1.__importDefault(require("clone"));
var method_1 = require("../utils/method");
var OrphanHttpServiceHandle_1 = require("./OrphanHttpServiceHandle");
var store_1 = require("../utils/store");
var RequestAction_1 = require("../../libs/RequestAction");
var instanceName_1 = require("../utils/instanceName");
var dev_1 = require("../utils/dev");
var BaseHttpService = /** @class */ (function () {
    function BaseHttpService(config) {
        this.caches = {};
        this.config = config;
    }
    BaseHttpService.prototype.action = function (fn) {
        var instanceName = instanceName_1.getInstanceName();
        if (!dev_1.useProxy()) {
            instanceName += '_' + instanceName_1.increaseActionCounter();
        }
        return new RequestAction_1.RequestAction({
            request: fn,
            instanceName: instanceName,
            runAction: this.withThrottle.bind(this),
            clearThrottle: this.clearThrottle.bind(this),
        });
    };
    BaseHttpService.prototype.getAsync = function (config) {
        var service = new OrphanHttpServiceHandle_1.OrphanHttpServiceHandle(config, method_1.METHOD.get, this);
        return this.withThrottle(service.collect());
    };
    BaseHttpService.prototype.postAsync = function (config) {
        var service = new OrphanHttpServiceHandle_1.OrphanHttpServiceHandle(config, method_1.METHOD.post, this);
        return this.withThrottle(service.collect());
    };
    BaseHttpService.prototype.putAsync = function (config) {
        var service = new OrphanHttpServiceHandle_1.OrphanHttpServiceHandle(config, method_1.METHOD.put, this);
        return this.withThrottle(service.collect());
    };
    BaseHttpService.prototype.deleteAsync = function (config) {
        var service = new OrphanHttpServiceHandle_1.OrphanHttpServiceHandle(config, method_1.METHOD.delete, this);
        return this.withThrottle(service.collect());
    };
    BaseHttpService.prototype.generateThrottleKey = function (action) {
        return JSON.stringify([
            action.reducerName,
            action.type,
            action.uri,
            action.method,
            action.body,
            action.query,
            action.requestOptions,
        ]);
    };
    BaseHttpService.prototype.clearThrottle = function (key) {
        delete this.caches[key];
    };
    BaseHttpService.prototype.withThrottle = function (action) {
        var _this = this;
        var throttleKey = this.generateThrottleKey(action);
        var key = action.type.success;
        var cacheData = this.caches[key];
        if (!action.useThrottle) {
            if (cacheData) {
                // Delete cache in case of toggle throttle
                delete cacheData[throttleKey];
            }
            return this.runAction(action);
        }
        var item = cacheData === null || cacheData === void 0 ? void 0 : cacheData[throttleKey];
        action.throttleKey = throttleKey;
        if (item) {
            if (Date.now() <= item.timestamp) {
                var promise = new Promise(function (resolve) {
                    var fakeAction = tslib_1.__assign(tslib_1.__assign({}, action), { type: action.type.success, response: clone_1.default(item.response, false), effect: action.onSuccess });
                    _this.next(fakeAction);
                    _this.triggerShowSuccess(fakeAction, action.successText);
                    resolve(fakeAction);
                });
                var wrapPromise = promise;
                wrapPromise.cancel = function () { };
                return wrapPromise;
            }
            else if (cacheData) {
                delete cacheData[action.throttleKey];
            }
        }
        return this.runAction(action);
    };
    BaseHttpService.prototype.collectResponse = function (action) {
        if (action.useThrottle && action.throttleMillSeconds > 0) {
            if (!this.caches[action.type]) {
                this.caches[action.type] = {};
            }
            this.caches[action.type][action.throttleKey] = {
                timestamp: Date.now() + action.throttleMillSeconds,
                response: clone_1.default(action.response, false),
            };
        }
    };
    BaseHttpService.prototype.next = function (action) {
        if (action.type) {
            store_1.getStore().dispatch(action);
        }
    };
    BaseHttpService.prototype.triggerShowSuccess = function (okResponse, successText) {
        if (successText) {
            this.config.onShowSuccess(successText, okResponse);
        }
    };
    BaseHttpService.prototype.triggerShowError = function (errorResponse, hideError) {
        if (!errorResponse.message) {
            return;
        }
        var showError;
        if (typeof hideError === 'boolean') {
            showError = !hideError;
        }
        else {
            showError = !hideError(errorResponse);
        }
        if (showError) {
            this.config.onShowError(errorResponse.message, errorResponse);
        }
    };
    return BaseHttpService;
}());
exports.BaseHttpService = BaseHttpService;
